from os import environ
import logging
from threading import Event, Lock
from time import sleep
from typing import Dict, Optional, Mapping, Tuple

from opentrons.drivers.utils import ParseError
from serial.serialutil import SerialException  # type: ignore

from opentrons.drivers import serial_communication, utils
from opentrons.drivers.serial_communication import SerialNoResponse

"""
- This driver is responsibe for providing an interface for the mag deck
- The driver is the only system component that knows about the mag-deck's
  GCODES or how the mag-deck communicates

- The driver is NOT responsible for interpreting deck states in any way
  or knowing anything about what the device is being used for
"""

log = logging.getLogger(__name__)

ERROR_KEYWORD = 'error'
ALARM_KEYWORD = 'alarm'

DEFAULT_MAG_DECK_TIMEOUT = 10   # Quite large to account for probe time

DEFAULT_STABILIZE_DELAY = 0.1
DEFAULT_COMMAND_RETRIES = 3

GCODES = {
    'HOME': 'G28.2',
    'PROBE_PLATE': 'G38.2',
    'GET_PLATE_HEIGHT': 'M836',
    'GET_CURRENT_POSITION': 'M114.2',
    'MOVE': 'G0',
    'DEVICE_INFO': 'M115',
    'PROGRAMMING_MODE': 'dfu'
}

MAG_DECK_BAUDRATE = 115200

MAG_DECK_COMMAND_TERMINATOR = '\r\n\r\n'
MAG_DECK_ACK = 'ok\r\nok\r\n'

MAG_DECK_MODELS = {
    'magneticModuleV1': 'mag_deck_v1.1',
    'magneticModuleV2': 'mag_deck_v20'
}

# Number of digits after the decimal point for millimeter values
# being sent to/from magnetic module
GCODE_ROUNDING_PRECISION = 3

mag_locks: Dict[str, Tuple[Lock, 'MagDeck']] = {}


class MagDeckError(Exception):
    pass


def _parse_distance_response(distance_string) -> float:
    """
    Parse responses of 'GET_PLATE_HEIGHT' & 'GET_CURRENT_POSITION'
    Example response of-
    GET_PLATE_HEIGHT: "height:12.34"
    GET_CURRENT_POSITION: "Z:12.34"
    """
    data = utils.parse_key_values(distance_string)
    val = data.get('Z', data.get('height'))
    if val is None:
        raise utils.ParseError(
            error_message='Unexpected argument to _parse_distance_response',
            parse_source=distance_string
        )

    return utils.parse_number(val, GCODE_ROUNDING_PRECISION)


class SimulatingDriver:
    def __init__(self, sim_model: str = None):
        self._port = None
        self._height = 0.0
        self._model = MAG_DECK_MODELS[sim_model] if sim_model\
            else 'mag_deck_v1.1'

    def probe_plate(self):
        pass

    def home(self):
        pass

    def move(self, location: float):
        self._height = location

    def get_device_info(self) -> Mapping[str, str]:
        return {'serial': 'dummySerialMD',
                'model': self._model,
                'version': 'dummyVersionMD'}

    def connect(self, port: str):
        pass

    def disconnect(self, port: str = None):
        pass

    def enter_programming_mode(self):
        pass

    @property
    def plate_height(self) -> float:
        return self._height

    @property
    def mag_position(self) -> float:
        return self._height

    def is_connected(self) -> bool:
        return True


class MagDeck:
    def __init__(self, config={}):
        self.run_flag = Event()
        self.run_flag.set()

        self._connection = None
        self._config = config

        self._plate_height: Optional[float] = None
        self._mag_position: Optional[float] = None
        self._port: Optional[str] = None
        self._lock: Optional[Lock] = None

    def connect(self, port=None) -> str:
        """
        :param port: '/dev/ot_module_magdeck[#]'
        NOTE: Using the symlink above to connect makes sure that the robot
        connects/reconnects to the module even after a device
        reset/reconnection
        """
        if environ.get('ENABLE_VIRTUAL_SMOOTHIE', '').lower() == 'true':
            return ''
        try:
            self.disconnect(port)
            self._connect_to_port(port)
            if mag_locks.get(port):
                self._lock = mag_locks[port][0]
            else:
                self._lock = Lock()
                mag_locks[port] = (self._lock, self)
            self._wait_for_ack()    # verify the device is there
            self._port = port

        except (SerialException, SerialNoResponse) as e:
            return str(e)
        return ''

    def disconnect(self, port=None):
        if port and self.is_connected():
            self._connection.close()  # type: ignore
            del mag_locks[port]
        elif self.is_connected():
            self._connection.close()  # type: ignore
        self._connection = None

    def is_connected(self) -> bool:
        # Does not detect if the module was physically plugged out
        # TODO: have it test actual connection
        if not self._connection:
            return False
        return self._connection.is_open

    @property
    def port(self) -> str:
        if not self._connection:
            return ''
        return self._connection.port

    def home(self) -> str:
        """
        Homes the magnet
        """
        self.run_flag.wait()
        try:
            self._send_command(GCODES['HOME'])
        except (MagDeckError, SerialException, SerialNoResponse) as e:
            return str(e)
        return ''

    def probe_plate(self) -> str:
        """
        Probes for the deck plate and calculates the plate distance
        from home.
        To be used for calibrating MagDeck
        """
        self.run_flag.wait()
        try:
            self._send_command(GCODES['PROBE_PLATE'])
        except (MagDeckError, SerialException, SerialNoResponse) as e:
            return str(e)
        return ''

    @property
    def plate_height(self) -> float:
        """
        Default plate_height for the device is 30;
        calculated as MAX_TRAVEL_DISTANCE(45mm) - 15mm
        """
        self._update_plate_height()
        assert self._plate_height is not None, 'not connected'
        return self._plate_height

    @property
    def mag_position(self) -> float:
        """
        Default mag_position for the device is 0.0
        i.e. it boots with the current position as 0.0
        """
        self._update_mag_position()
        assert self._mag_position is not None, 'not connected'
        return self._mag_position

    def move(self, position_mm) -> str:
        """
        Move the magnets along Z axis where the home position is 0.0;
        position_mm-> a point along Z. Does not self-check if the position
        is outside of the deck's linear range
        """
        self.run_flag.wait()
        try:
            position_mm = round(
                float(position_mm), GCODE_ROUNDING_PRECISION)
            self._send_command(
                '{0} Z{1}'.format(GCODES['MOVE'], position_mm))
        except (MagDeckError, SerialException, SerialNoResponse) as e:
            return str(e)
        return ''

    def get_device_info(self) -> Dict[str, str]:
        """
        Queries Temp-Deck for it's build version, model, and serial number

        returns: dict
            Where keys are the strings 'version', 'model', and 'serial',
            and each value is a string identifier

            {
                'serial': '1aa11bb22',
                'model': '1aa11bb22',
                'version': '1aa11bb22'
            }

        Example input from Temp-Deck's serial response:
            "serial:aa11bb22 model:aa11bb22 version:aa11bb22"
        """
        return self._recursive_get_info(DEFAULT_COMMAND_RETRIES)

    def enter_programming_mode(self) -> str:
        """
        Enters and stays in DFU mode for 8 seconds.
        The module resets upon exiting the mode
        which causes the robot to lose serial connection to it.
        The connection can be restored by performing a .disconnect()
        followed by a .connect() to the same symlink node
        """
        try:
            self._send_command(GCODES['PROGRAMMING_MODE'])
        except (MagDeckError, SerialException, SerialNoResponse) as e:
            return str(e)
        if self._port:
            del mag_locks[self._port]
        return ''

    def _recursive_write_and_return(self, cmd, timeout, retries, tag=None):
        if not tag:
            tag = f'magdeck {id(self)}'
        try:
            return serial_communication.write_and_return(
                cmd,
                MAG_DECK_ACK,
                self._connection,
                timeout,
                tag=tag)
        except SerialNoResponse as e:
            retries -= 1
            if retries <= 0:
                raise e
            sleep(DEFAULT_STABILIZE_DELAY)
            if self._connection:
                self._connection.close()
                self._connection.open()
            return self._recursive_write_and_return(
                cmd, timeout, retries, tag=tag)

    def _wait_for_ack(self):
        """
        This methods writes a sequence of newline characters, which will
        guarantee mag-deck responds with 'ok\r\nok\r\n' within 1 seconds
        """
        self._send_command('\r\n', timeout=DEFAULT_MAG_DECK_TIMEOUT)

    # Potential place for command optimization (buffering, flushing, etc)
    def _send_command(self, command, timeout=DEFAULT_MAG_DECK_TIMEOUT):
        command_line = command + ' ' + MAG_DECK_COMMAND_TERMINATOR
        assert self._lock, 'need a lock'
        with self._lock:
            ret_code = self._recursive_write_and_return(
                command_line, timeout, DEFAULT_COMMAND_RETRIES)

            # Smoothieware returns error state if a switch was hit while moving
            if (ERROR_KEYWORD in ret_code.lower()) or \
                    (ALARM_KEYWORD in ret_code.lower()):
                log.error(f'Received error message from Mag-Deck: {ret_code}')
                raise MagDeckError(ret_code)

            return ret_code.strip()

    def _connect_to_port(self, port=None):
        try:
            self._connection = serial_communication.connect(
                port=port,
                baudrate=MAG_DECK_BAUDRATE
            )
        except SerialException:
            # if another process is using the port, pyserial raises an
            # exception that describes a "readiness to read" which is confusing
            error_msg = "Unable to access Serial port to Mag-Deck. This is"
            " because another process is currently using it, or"
            " the Serial port is disabled on this device (OS)"
            raise SerialException(error_msg)
        except TypeError:
            # This happens if there are no ot_module_magdeck* devices in /dev
            # For development use ENABLE_VIRTUAL_SMOOTHIE=true
            raise SerialException('No port specified')

    def _update_plate_height(self) -> str:
        try:
            res = self._send_command(GCODES['GET_PLATE_HEIGHT'])
            distance = _parse_distance_response(res)
        except (MagDeckError, SerialException, SerialNoResponse) as e:
            return str(e)
        self._plate_height = distance
        return ''

    def _update_mag_position(self) -> str:
        try:
            res = self._send_command(GCODES['GET_CURRENT_POSITION'])
            distance = _parse_distance_response(res)
        except (MagDeckError, SerialException, SerialNoResponse) as e:
            return str(e)
        self._mag_position = distance
        return ''

    def _recursive_get_info(self, retries) -> Dict[str, str]:
        try:
            device_info = self._send_command(GCODES['DEVICE_INFO'])
            return utils.parse_device_information(device_info)
        except ParseError as e:
            retries -= 1
            if retries <= 0:
                raise MagDeckError(e)
            sleep(DEFAULT_STABILIZE_DELAY)
            return self._recursive_get_info(retries)



""" Type definitions for modules in this tree """
from dataclasses import dataclass
from typing import Dict, NamedTuple, Optional
from enum import Enum


class MoveSplit(NamedTuple):
    split_distance: float
    split_current: float
    split_speed: float
    after_time: float
    fullstep: bool


MoveSplits = Dict[str, MoveSplit]
#: Dict mapping axes to their split parameters


@dataclass
class Temperature:
    """Tempdeck temperature and thermocycler plate temperature."""
    current: Optional[float]
    target: Optional[float]


@dataclass
class PlateTemperature(Temperature):
    """Thermocycler lid temperature"""
    hold: Optional[float]


class ThermocyclerLidStatus(str, Enum):
    """Thermocycler lid status."""
    UNKNOWN = 'unknown'
    CLOSED = 'closed'
    IN_BETWEEN = 'in_between'
    OPEN = 'open'
    MAX = 'max'



"""Aspirate command request, result, and implementation models."""
from __future__ import annotations
from typing_extensions import final

from .command import CommandImplementation, CommandHandlers
from .pipetting_common import BaseLiquidHandlingRequest, BaseLiquidHandlingResult


@final
class AspirateRequest(BaseLiquidHandlingRequest):
    """A request to move to a specific well and aspirate from it."""

    def get_implementation(self) -> AspirateImplementation:
        """Get the execution implementation of the AspirateRequest."""
        return AspirateImplementation(self)


@final
class AspirateResult(BaseLiquidHandlingResult):
    """Result data from the execution of a AspirateRequest."""

    pass


@final
class AspirateImplementation(
    CommandImplementation[AspirateRequest, AspirateResult]
):
    """Aspirate command implementation."""

    async def execute(self, handlers: CommandHandlers) -> AspirateResult:
        """Move to and aspirate from the requested well."""
        volume = await handlers.pipetting.aspirate(
            pipette_id=self._request.pipetteId,
            labware_id=self._request.labwareId,
            well_name=self._request.wellName,
            well_location=self._request.wellLocation,
            volume=self._request.volume,
        )

        return AspirateResult(volume=volume)



"""An emulation of the opentrons thermocycler module.

The purpose is to provide a fake backend that responds to GCODE commands.
"""

import logging
from typing import Optional
from opentrons.drivers.thermocycler.driver import GCODES
from opentrons.drivers.types import ThermocyclerLidStatus
from opentrons.hardware_control.emulation.parser import Parser, Command

from .abstract_emulator import AbstractEmulator
from . import util

logger = logging.getLogger(__name__)

GCODE_OPEN_LID = GCODES['OPEN_LID']
GCODE_CLOSE_LID = GCODES['CLOSE_LID']
GCODE_GET_LID_STATUS = GCODES['GET_LID_STATUS']
GCODE_SET_LID_TEMP = GCODES['SET_LID_TEMP']
GCODE_GET_LID_TEMP = GCODES['GET_LID_TEMP']
GCODE_EDIT_PID_PARAMS = GCODES['EDIT_PID_PARAMS']
GCODE_SET_PLATE_TEMP = GCODES['SET_PLATE_TEMP']
GCODE_GET_PLATE_TEMP = GCODES['GET_PLATE_TEMP']
GCODE_SET_RAMP_RATE = GCODES['SET_RAMP_RATE']
GCODE_DEACTIVATE_ALL = GCODES['DEACTIVATE_ALL']
GCODE_DEACTIVATE_LID = GCODES['DEACTIVATE_LID']
GCODE_DEACTIVATE_BLOCK = GCODES['DEACTIVATE_BLOCK']
GCODE_DEVICE_INFO = GCODES['DEVICE_INFO']

SERIAL = "thermocycler_emulator"
MODEL = "v02"
VERSION = "v1.1.0"


class ThermocyclerEmulator(AbstractEmulator):
    """Thermocycler emulator"""

    def __init__(self, parser: Parser) -> None:
        self.lid_target_temp = util.OptionalValue[float]()
        self.lid_current_temp: float = util.TEMPERATURE_ROOM
        self.lid_status = ThermocyclerLidStatus.CLOSED
        self.lid_at_target: Optional[bool] = None
        self.plate_total_hold_time = util.OptionalValue[float]()
        self.plate_time_remaining = util.OptionalValue[float]()
        self.plate_target_temp = util.OptionalValue[float]()
        self.plate_current_temp: float = util.TEMPERATURE_ROOM
        self.plate_volume = util.OptionalValue[float]()
        self.plate_at_target = util.OptionalValue[float]()
        self.plate_ramp_rate = util.OptionalValue[float]()
        self._parser = parser

    def handle(self, line: str) -> Optional[str]:
        """Handle a line"""
        results = (self._handle(c) for c in self._parser.parse(line))
        joined = ' '.join(r for r in results if r)
        return None if not joined else joined

    def _handle(self, command: Command) -> Optional[str]:  # noqa: C901
        """
        Handle a command.

        TODO: AL 20210218 create dispatch map and remove 'noqa(C901)'
        """
        logger.info(f"Got command {command}")
        if command.gcode == GCODE_OPEN_LID:
            self.lid_status = ThermocyclerLidStatus.OPEN
        elif command.gcode == GCODE_CLOSE_LID:
            self.lid_status = ThermocyclerLidStatus.CLOSED
        elif command.gcode == GCODE_GET_LID_STATUS:
            return f"Lid:{self.lid_status}"
        elif command.gcode == GCODE_SET_LID_TEMP:
            temperature = command.params['S']
            assert isinstance(temperature, float),\
                f"invalid temperature '{temperature}'"
            self.lid_target_temp.val = temperature
            self.lid_current_temp = self.lid_target_temp.val
        elif command.gcode == GCODE_GET_LID_TEMP:
            return f"T:{self.lid_target_temp} C:{self.lid_current_temp} " \
                   f"H:none Total_H:none At_target?:0"
        elif command.gcode == GCODE_EDIT_PID_PARAMS:
            pass
        elif command.gcode == GCODE_SET_PLATE_TEMP:
            for prefix, value in command.params.items():
                assert isinstance(value, float), f"invalid value '{value}'"
                if prefix == 'S':
                    self.plate_target_temp.val = value
                    self.plate_current_temp = self.plate_target_temp.val
                elif prefix == 'V':
                    self.plate_volume.val = value
                elif prefix == 'H':
                    self.plate_total_hold_time.val = value
                    self.plate_time_remaining.val = value
        elif command.gcode == GCODE_GET_PLATE_TEMP:
            return f"T:{self.plate_target_temp} " \
                   f"C:{self.plate_current_temp} " \
                   f"H:{self.plate_time_remaining} " \
                   f"Total_H:{self.plate_total_hold_time} " \
                   f"At_target?:{self.plate_at_target}"
        elif command.gcode == GCODE_SET_RAMP_RATE:
            self.plate_ramp_rate.val = command.params['S']
        elif command.gcode == GCODE_DEACTIVATE_ALL:
            pass
        elif command.gcode == GCODE_DEACTIVATE_LID:
            pass
        elif command.gcode == GCODE_DEACTIVATE_BLOCK:
            pass
        elif command.gcode == GCODE_DEVICE_INFO:
            return f"serial:{SERIAL} model:{MODEL} version:{VERSION}"
        return None

    @staticmethod
    def get_terminator() -> bytes:
        return b'\r\n'



from enum import Enum
from dataclasses import dataclass
from typing import (
    Dict, NamedTuple, Callable, Any, Type, TypeVar,
    Tuple, Awaitable, Mapping, Union, TYPE_CHECKING)
from pathlib import Path

from opentrons.drivers.rpi_drivers.types import USBPort

if TYPE_CHECKING:
    from opentrons_shared_data.module.dev_types import (
        ThermocyclerModuleType, MagneticModuleType, TemperatureModuleType)

ThermocyclerStep = Dict[str, float]

InterruptCallback = Callable[[str], None]

UploadFunction = Callable[[str, str, Dict[str, Any]],
                          Awaitable[Tuple[bool, str]]]

LiveData = Mapping[str, Union[str, Mapping[str, Union[float, str, None]]]]


E = TypeVar('E', bound='_ProvideLookup')


class _ProvideLookup(Enum):
    @classmethod
    def from_str(cls: Type[E], typename: str) -> 'E':
        for m in cls.__members__.values():
            if m.value == typename:
                return m
        raise AttributeError(f'No such type {typename}')


class ModuleType(_ProvideLookup):
    THERMOCYCLER: 'ThermocyclerModuleType' = 'thermocyclerModuleType'
    TEMPERATURE: 'TemperatureModuleType' = 'temperatureModuleType'
    MAGNETIC: 'MagneticModuleType' = 'magneticModuleType'


class MagneticModuleModel(_ProvideLookup):
    MAGNETIC_V1: str = 'magneticModuleV1'
    MAGNETIC_V2: str = 'magneticModuleV2'


class TemperatureModuleModel(_ProvideLookup):
    TEMPERATURE_V1: str = 'temperatureModuleV1'
    TEMPERATURE_V2: str = 'temperatureModuleV2'


class ThermocyclerModuleModel(_ProvideLookup):

    @classmethod
    def from_str(cls: Type['ThermocyclerModuleModel'],
                 typename: str) -> 'ThermocyclerModuleModel':
        return super().from_str(typename)

    THERMOCYCLER_V1: str = 'thermocyclerModuleV1'


@dataclass
class ModuleAtPort:
    port: str
    name: str
    usb_port: USBPort = USBPort(name='', sub_names=[])


class BundledFirmware(NamedTuple):
    """ Represents a versioned firmware file, generally bundled into the fs"""
    version: str
    path: Path

    def __repr__(self) -> str:
        return f'<BundledFirmware {self.version}, path={self.path}>'


class UpdateError(RuntimeError):
    pass


class UnsupportedModuleError(Exception):
    pass


class AbsentModuleError(Exception):
    pass


class ModuleInfo(NamedTuple):
    model: str        # A module model such as "magneticModuleV2"
    fw_version: str   # The version of the firmware
    hw_revision: str  # the revision of the hardware
    serial: str       # the serial number


ModuleModel = Union[
    MagneticModuleModel, TemperatureModuleModel, ThermocyclerModuleModel]



""" Utility functions and classes for the hardware controller"""
import asyncio
import logging
from enum import Enum
from typing import Dict, Any, Optional, List, Mapping, Tuple

from .types import CriticalPoint, MotionChecks, OutOfBoundsMove, Axis
from opentrons.types import Point

mod_log = logging.getLogger(__name__)


def _handle_loop_exception(loop: asyncio.AbstractEventLoop,
                           context: Dict[str, Any]):
    mod_log.error(f"Caught exception: {context}")


def use_or_initialize_loop(loop: Optional[asyncio.AbstractEventLoop]
                           ) -> asyncio.AbstractEventLoop:
    checked_loop = loop or asyncio.get_event_loop()
    checked_loop.set_exception_handler(_handle_loop_exception)
    return checked_loop


def plan_arc(
        origin_point: Point,
        dest_point: Point,
        z_height: float,
        origin_cp: CriticalPoint = None,
        dest_cp: CriticalPoint = None,
        extra_waypoints: List[Tuple[float, float]] = None)\
        -> List[Tuple[Point, Optional[CriticalPoint]]]:

    assert z_height >= max(origin_point.z, dest_point.z)
    checked_wp = extra_waypoints or []
    return [(origin_point._replace(z=z_height), origin_cp)]\
        + [(Point(x=wp[0], y=wp[1], z=z_height), dest_cp)
           for wp in checked_wp]\
        + [(dest_point._replace(z=z_height), dest_cp),
           (dest_point, dest_cp)]


class DeckTransformState(Enum):
    OK = "OK"
    IDENTITY = "IDENTITY"
    BAD_CALIBRATION = "BAD_CALIBRATION"
    SINGULARITY = "SINGULARITY"

    def __str__(self):
        return self.name


def check_motion_bounds(
        target_smoothie: Mapping[Axis, float],
        target_deck: Mapping[Axis, float],
        bounds: Mapping[Axis, Tuple[float, float]],
        checks: MotionChecks):
    """
    Log or raise an error (depending on checks) if a specified
    target position is outside the acceptable bounds of motion

    Which axes are checked is defined by the elements of
    target_smoothie.

    The target_deck and mapping does not need to be
    entirely filled; it is used only for logging
    """
    bounds_message_format = (
        'Out of bounds move: {axis}=({tsp} motor controller, {tdp} deck) too '
        '{dir} for limit {limsp}')
    for ax in target_smoothie.keys():
        if target_smoothie[ax] < bounds[ax][0]:
            bounds_message = bounds_message_format.format(
                axis=ax,
                tsp=target_smoothie[ax],
                tdp=target_deck.get(ax, 'unknown'),
                dir='low',
                limsp=bounds.get(ax, ('unknown',))[0],
            )
            mod_log.warning(bounds_message)
            if checks.value & MotionChecks.LOW.value:
                raise OutOfBoundsMove(bounds_message)
        elif target_smoothie[ax] > bounds[ax][1]:
            bounds_message = bounds_message_format.format(
                axis=ax,
                tsp=target_smoothie[ax],
                tdp=target_deck.get(ax, 'unknown'),
                dir='high',
                limsp=bounds.get(ax, (None, 'unknown'))[1],
            )
            mod_log.warning(bounds_message)
            if checks.value & MotionChecks.HIGH.value:
                raise OutOfBoundsMove(bounds_message)



from .mod_abc import AbstractModule
from .tempdeck import TempDeck
from .magdeck import MagDeck
from .thermocycler import Thermocycler
from .update import update_firmware
from .utils import MODULE_HW_BY_NAME, build
from .types import (ThermocyclerStep, InterruptCallback, UploadFunction,
                    BundledFirmware, UpdateError, UnsupportedModuleError,
                    AbsentModuleError, ModuleAtPort)

__all__ = [
    'MODULE_HW_BY_NAME', 'build', 'update_firmware',
    'ThermocyclerStep', 'AbstractModule', 'TempDeck',
    'MagDeck', 'Thermocycler', 'InterruptCallback',
    'UploadFunction', 'BundledFirmware', 'UpdateError',
    'UnsupportedModuleError', 'AbsentModuleError', 'ModuleAtPort'
]



import logging
from typing import Dict, List, TYPE_CHECKING, Union
from opentrons.protocol_api.contexts import ProtocolContext, \
    MagneticModuleContext, TemperatureModuleContext, ModuleContext, \
    ThermocyclerContext
from .execute_json_v3 import _delay, _move_to_slot
from opentrons.protocols.execution.types import LoadedLabware, Instruments
from opentrons_shared_data.protocol.constants import (
    JsonRobotCommand, JsonPipetteCommand)


if TYPE_CHECKING:
    from opentrons_shared_data.protocol.dev_types import (
        JsonProtocolV4,
        JsonProtocolV5,
        MagneticModuleEngageParams,
        ModuleIDParams, TemperatureParams,
        ThermocyclerSetTargetBlockParams,
        ThermocyclerRunProfileParams, Command,
        TemperatureModuleCommandId, MagneticModuleCommandId,
        ThermocyclerCommandId
    )
    from opentrons.protocols.execution.dev_types import (
        PipetteDispatch, JsonV4MagneticModuleDispatch,
        JsonV4TemperatureModuleDispatch,
        JsonV4ThermocyclerDispatch)

MODULE_LOG = logging.getLogger(__name__)

# Special string for slot used by thermocycler in JSON protocols.
TC_SPANNING_SLOT = 'span7_8_10_11'


def load_labware_from_json_defs(
        ctx: ProtocolContext,
        protocol: 'JsonProtocolV4',
        modules: Dict[str, ModuleContext]) -> LoadedLabware:
    protocol_labware = protocol['labware']
    definitions = protocol['labwareDefinitions']
    loaded_labware = {}

    for labware_id, props in protocol_labware.items():
        slot = props['slot']
        definition = definitions[props['definitionId']]
        label = props.get('displayName', None)
        if slot in modules:
            loaded_labware[labware_id] = \
                modules[slot].load_labware_from_definition(definition, label)
        else:
            loaded_labware[labware_id] = \
                ctx.load_labware_from_definition(definition, slot, label)

    return loaded_labware


def load_modules_from_json(
        ctx: ProtocolContext,
        protocol: 'JsonProtocolV4') -> Dict[str, ModuleContext]:
    module_data = protocol['modules']
    modules_by_id = {}

    # the sort order doesn't matter, we just need it to be stable
    # to ensure `load_module` side-effects are deterministic
    # (eg, multiple module support)
    sorted_modules = sorted(
        module_data.items(),
        key=lambda v: v[1]['slot'] +
        v[1]['model'])

    for module_id, module_props in sorted_modules:
        model = module_props['model']
        slot = module_props['slot']
        if slot == TC_SPANNING_SLOT:
            instr = ctx.load_module(model)
        else:
            instr = ctx.load_module(model, slot)
        modules_by_id[module_id] = instr

    return modules_by_id


def _engage_magnet(
        module: MagneticModuleContext,
        params: 'MagneticModuleEngageParams') -> None:
    engage_height = params['engageHeight']
    module.engage(height_from_base=engage_height)


def _disengage_magnet(
        module: MagneticModuleContext,
        params: 'ModuleIDParams') -> None:
    module.disengage()


def _temperature_module_set_temp(module: TemperatureModuleContext,
                                 params: 'TemperatureParams') -> None:
    temperature = params['temperature']
    module.start_set_temperature(temperature)


def _temperature_module_deactivate(module: TemperatureModuleContext,
                                   params: 'ModuleIDParams') -> None:
    module.deactivate()


def _temperature_module_await_temp(module: TemperatureModuleContext,
                                   params: 'TemperatureParams') -> None:
    temperature = params['temperature']
    module.await_temperature(temperature)


def _thermocycler_close_lid(module: ThermocyclerContext,
                            params: 'ModuleIDParams') -> None:
    module.close_lid()


def _thermocycler_open_lid(module: ThermocyclerContext,
                           params: 'ModuleIDParams') -> None:
    module.open_lid()


def _thermocycler_deactivate_block(module: ThermocyclerContext,
                                   params: 'ModuleIDParams') -> None:
    module.deactivate_block()


def _thermocycler_deactivate_lid(module: ThermocyclerContext,
                                 params: 'ModuleIDParams') -> None:
    module.deactivate_lid()


def _thermocycler_set_block_temperature(
        module: ThermocyclerContext,
        params: 'ThermocyclerSetTargetBlockParams') -> None:
    temperature = params['temperature']
    module.set_block_temperature(temperature)


def _thermocycler_set_lid_temperature(
        module: ThermocyclerContext,
        params: 'TemperatureParams') -> None:
    temperature = params['temperature']
    module.set_lid_temperature(temperature)


def _thermocycler_run_profile(
        module: ThermocyclerContext,
        params: 'ThermocyclerRunProfileParams') -> None:
    volume = params['volume']
    profile = [{
        'temperature': p['temperature'],
        'hold_time_seconds': p['holdTime']
    } for p in params['profile']]
    module.execute_profile(
        steps=profile,
        block_max_volume=volume,
        repetitions=1)


def assert_no_async_tc_behavior(commands: List['Command']) -> None:
    # awaiters[i] corresponds to setters[i]
    setters = [
        "thermocycler/setTargetBlockTemperature",
        "thermocycler/setTargetLidTemperature",
        "thermocycler/runProfile"]
    awaiters = [
        "thermocycler/awaitBlockTemperature",
        "thermocycler/awaitLidTemperature",
        "thermocycler/awaitProfileComplete"]

    command_types = [c['command'] for c in commands]

    first_command = command_types[0]
    if first_command in awaiters:
        raise RuntimeError(
            ("Unsupported behavior. Cannot {} as the " +
             "first command of a protocol")
            .format(first_command))

    last_command = command_types[-1]
    if last_command in setters:
        raise RuntimeError(
            ("Unsupported behavior. Cannot {} as the " +
             "last command of a protocol")
            .format(last_command))

    # [a, b, c, d] -> [a, b], [b, c], [c, d]
    for a, b in zip(command_types, command_types[1:]):
        #  a setter must be immediately followed
        # by its corresponding awaiter
        if a in setters:
            expected_awaiter = awaiters[setters.index(a)]
            if b != expected_awaiter:
                raise RuntimeError(
                    ("Unsupported behavior. {} must be immediately" +
                     "followed by {}")
                    .format(a, expected_awaiter))
        # an awaiter must be preceded by a setter
        elif b in awaiters:
            expected_setter = setters[awaiters.index(b)]
            raise RuntimeError(
                ("Unsupported behavior. {} must be immediately " +
                 "preceded by {}").format(b, expected_setter))


def assert_tc_commands_do_not_use_unimplemented_params(
        commands: List['Command']) -> None:
    # raise errors if commands include optional param keys that
    # the executor/api does not currently support
    unsupported_keys_by_command = {
        "thermocycler/setTargetBlockTemperature": ["volume"]
    }
    for c in commands:
        command_type = c['command']
        params = c['params']
        unsupported_keys = unsupported_keys_by_command.get(
            command_type, [])
        for k in unsupported_keys:
            if k in params:
                raise RuntimeError((
                    "{} does not support {} param. " +
                    "This may be implemented in a later version of " +
                    "the robot server").format(command_type, k))


def dispatch_json(
        context: ProtocolContext,
        protocol_data: Union['JsonProtocolV4', 'JsonProtocolV5'],
        instruments: Instruments,
        loaded_labware: LoadedLabware,
        modules: Dict[str, ModuleContext],
        pipette_command_map: 'PipetteDispatch',
        magnetic_module_command_map: 'JsonV4MagneticModuleDispatch',
        temperature_module_command_map: 'JsonV4TemperatureModuleDispatch',
        thermocycler_module_command_map: 'JsonV4ThermocyclerDispatch'
) -> None:
    commands = protocol_data['commands']

    assert_no_async_tc_behavior(commands)
    assert_tc_commands_do_not_use_unimplemented_params(commands)

    for command_item in commands:
        command_type = command_item['command']
        params = command_item['params']
        # because of https://github.com/python/mypy/issues/8940
        # we can't narrow down types using in sadly
        if command_type in pipette_command_map:
            pipette_command_map[command_type](  # type: ignore
                instruments, loaded_labware, params)
        elif command_type in magnetic_module_command_map:
            handleMagnetCommand(
                params,  # type: ignore
                modules,
                command_type,  # type: ignore
                magnetic_module_command_map
            )
        elif command_type in temperature_module_command_map:
            handleTemperatureCommand(params,  # type: ignore
                                     modules,
                                     command_type,  # type: ignore
                                     temperature_module_command_map)
        elif command_type in thermocycler_module_command_map:
            handleThermocyclerCommand(params,  # type: ignore
                                      modules,  # type: ignore
                                      command_type,  # type: ignore
                                      thermocycler_module_command_map)
        elif command_item['command'] == JsonRobotCommand.delay.value:
            _delay(context, params)  # type: ignore
        elif command_type == JsonPipetteCommand.moveToSlot.value:
            _move_to_slot(context, instruments, params)  # type: ignore
        else:
            raise RuntimeError(
                "Unsupported command type {}".format(command_type))


def handleTemperatureCommand(
        params: Union['TemperatureParams', 'ModuleIDParams'],
        modules: Dict[str, ModuleContext],
        command_type: 'TemperatureModuleCommandId',
        temperature_module_command_map
) -> None:
    module_id = params['module']
    module = modules[module_id]
    if isinstance(module, TemperatureModuleContext):
        temperature_module_command_map[command_type](
            module, params
        )
    else:
        raise RuntimeError(
            "Temperature Module does not match " +
            "TemperatureModuleContext interface"
        )


def handleThermocyclerCommand(
        params: Union['ModuleIDParams', 'TemperatureParams',
                      'ThermocyclerRunProfileParams',
                      'ThermocyclerSetTargetBlockParams'],
        modules: Dict[str, ThermocyclerContext],
        command_type: 'ThermocyclerCommandId',
        thermocycler_module_command_map
) -> None:
    module_id = params['module']
    module = modules[module_id]
    if isinstance(module, ThermocyclerContext):
        thermocycler_module_command_map[command_type](
            module, params
        )
    else:
        raise RuntimeError(
            "Thermocycler Module does not match ThermocyclerContext interface"
        )


def handleMagnetCommand(
        params: Union['ModuleIDParams', 'MagneticModuleEngageParams'],
        modules: Dict[str, ModuleContext],
        command_type: 'MagneticModuleCommandId',
        magnetic_module_command_map
) -> None:
    module_id = params['module']
    module = modules[module_id]
    if isinstance(module, MagneticModuleContext):
        magnetic_module_command_map[command_type](
            module, params
        )
    else:
        raise RuntimeError(
            "Magnetic Module does not match MagneticModuleContext interface"
        )



# noqa: D100
from __future__ import annotations
from typing import Any, List, Dict, Optional, Union, cast

from opentrons.protocol_engine.clients import SyncClient as ProtocolEngineClient
from .errors import LabwareIsNotTipRackError
from .types import DeckSlotName, LabwareParameters, Point
from .well import Well


class Labware:  # noqa: D101
    def __init__(
        self,
        engine_client: ProtocolEngineClient,
        labware_id: str,
    ) -> None:
        """Initialize a Labware API provider.

        You should not need to call this constructor yourself. The system will
        create a `Labware` for you when you call :py:meth:`load_labware`.

        Args:
            engine_client: A client to access protocol state.
            labware_id: The labware's identifier in commands and protocol state.
        """
        self._engine_client = engine_client
        self._labware_id = labware_id

    # TODO(mc, 2021-04-22): remove this property; it's redundant and
    # unlikely to be used by PAPI users
    @property
    def api_version(self) -> Any:  # noqa: D102
        raise NotImplementedError()

    @property
    def labware_id(self) -> str:
        """Unique identifier for this labware instance in the protocol.

        This identifier is used to reference this labware in commands and
        protocol state. This ID will be unique for every piece of labware
        in the protocol, even if labware share the same definition.
        """
        return self._labware_id

    @property
    def uri(self) -> str:
        """A string fully identifying the labware's underlying definition.

        The labware's definition URI will be distinct and different than the
        labware's ``labware_id``, which refers instead to specific instance of
        the labware on the deck during the protocol.

        The definition URI is of the format ``"{namespace}/{load_name}/{version}"``.
        """
        return self._engine_client.state.labware.get_definition_uri(
            labware_id=self._labware_id
        )

    # TODO(mc, 2021-04-22): labware may be on a module, replace Any with Module
    @property
    def parent(self) -> Union[DeckSlotName, Any]:
        """The parent location of this labware.

        If the labware's parent is a string, that string represents a specific
        deck slot. Otherwise, the labware is on a module, and the object
        returned will be a :py:class:`ModuleContext`.
        """
        parent = self._engine_client.state.labware.get_labware_location(
            labware_id=self._labware_id
        )
        return str(parent.slot)

    # TODO(mc, 2021-05-03): document removal of name setter
    @property
    def name(self) -> str:  # noqa: D102
        raise NotImplementedError()

    @property
    def load_name(self) -> str:
        """The load name of this labware's definition.

        This is best used for informational purposes. There is no requirement
        that a load name be unique among other definitions. To specify a fully-
        qualified definition, use :py:meth:`uri`.
        """
        return self._engine_client.state.labware.get_load_name(
            labware_id=self._labware_id
        )

    # TODO(mc, 2021-05-03): this is an internal dictionary of Opentrons-specific
    # data; does it really need to be a public property? Can we expose the
    # definition and call it a day?
    @property
    def parameters(self) -> LabwareParameters:  # noqa: D102
        definition = self._engine_client.state.labware.get_labware_definition(
            labware_id=self._labware_id
        )
        return cast(LabwareParameters, definition.parameters.dict())

    # TODO(mc, 2021-05-03): this is an internal list of Opentrons-specific
    # data; does it really need to be a public property? Can we expose the
    # definition and call it a day?
    @property
    def quirks(self) -> List[str]:  # noqa: D102
        return self._engine_client.state.labware.get_quirks(labware_id=self._labware_id)

    # TODO(mc, 2021-05-03): this property appears to be primarily for magdeck
    # operational logic, and its presence in this interface is no longer
    # necessary with Protocol Engine controlling execution. Can we get rid of it?
    @property
    def magdeck_engage_height(self) -> Optional[float]:    # noqa: D102
        definition = self._engine_client.state.labware.get_labware_definition(
            labware_id=self._labware_id
        )
        return definition.parameters.magneticModuleEngageHeight

    @property
    def calibrated_offset(self) -> Point:
        """The location of the labware's front-bottom-left corner in deck coordinates.

        This value takes into account the labware's calibration data.
        """
        return self._engine_client.state.geometry.get_labware_position(
            labware_id=self._labware_id
        )

    @property
    def highest_z(self) -> float:
        """The z-coordinate of the highest point on the labware.

        This value takes into account the labware's definition as well
        as its calibration data.
        """
        return self._engine_client.state.geometry.get_labware_highest_z(
            labware_id=self._labware_id
        )

    # TODO(mc, 2021-05-03): encode this in a specific `TipRack` interface that
    # extends from Labware
    @property
    def is_tiprack(self) -> bool:
        """Whether this labware is a tiprack."""
        definition = self._engine_client.state.labware.get_labware_definition(
            labware_id=self._labware_id
        )
        return definition.parameters.isTiprack

    # TODO(mc, 2021-05-03): encode this in a specific `TipRack` interface that
    # extends from Labware
    # TODO(mc, 2021-05-03): this property appears to be primarily for pipette
    # operational logic, and its presence in this interface is no longer
    # necessary with Protocol Engine controlling execution. Does it need to be
    # public? Can we expose the definition and call it a day?
    # TODO(mc, 2021-05-03): document removal of tip_length setter
    @property
    def tip_length(self) -> float:
        """The nominal length of tips if this labware is a tiprack.

        Raises:
            LabwareIsNotTipRackError: will raise if this property is accessed
                on a labware instance that is not a tip rack.
        """
        definition = self._engine_client.state.labware.get_labware_definition(
            labware_id=self._labware_id
        )
        if definition.parameters.tipLength is None:
            raise LabwareIsNotTipRackError(f"{self.load_name} is not a tip rack.")
        return definition.parameters.tipLength

    def well(self, idx: int) -> Well:  # noqa: D102
        raise NotImplementedError()

    def wells(self) -> List[Well]:  # noqa: D102
        raise NotImplementedError()

    def wells_by_name(self) -> Dict[str, Well]:  # noqa: D102
        raise NotImplementedError()

    def rows(self) -> List[List[Well]]:  # noqa: D102
        raise NotImplementedError()

    def rows_by_name(self) -> Dict[str, List[Well]]:  # noqa: D102
        raise NotImplementedError()

    def columns(self) -> List[List[Well]]:  # noqa: D102
        raise NotImplementedError()

    def columns_by_name(self) -> Dict[str, List[Well]]:  # noqa: D102
        raise NotImplementedError()

    def __repr__(self) -> str:  # noqa: D105
        raise NotImplementedError()

    def __eq__(self, other: object) -> bool:
        """Compare for object equality.

        Checks that other object is a `Labware` and has the same identifier.
        """
        return isinstance(other, Labware) and self._labware_id == other._labware_id

    def __hash__(self) -> int:
        """Get hash.

        Uses the labware instance's unique identifier in protocol state.
        """
        return hash(self._labware_id)

    def __getitem__(self, key: str) -> Well:  # noqa: D105
        raise NotImplementedError()

    # todo(mm, 2021-04-09): The following methods appear on docs.opentrons.com
    # (accidentally?) but aren't versioned. Figure out whether we need to
    # include them here.
    #   * next_tip()
    #   * use_tips()
    #   * previous_tip()
    #   * return_tips()


# todo(mm, 2021-04-09): In addition to the Labware class, the APIv2
# analogue to this module provides several free functions. Some of them
# appear on docs.opentrons.com, but none of them are versioned with
# @requires_version, so it's unclear if they're meant to be part of the public
# API. We need to figure out whether we need to include them here.
# todo(MC, 2021-05-04): In support of those free functions, there are two
# errors defined: `TipSelectionError` and `OutOfTipsError`. We need to figure
# out if we need them, too.



"""CommandQueueWorker definition. Implements JSON protocol flow control."""


class CommandQueueWorker:
    """A class that executes the queued commands in a ProtocolEngine."""

    def play(self) -> None:
        """Start/resume queued command execution."""
        raise NotImplementedError()

    def pause(self) -> None:
        """Pause queued command execution."""
        raise NotImplementedError()

    def stop(self) -> None:
        """Stop queued command execution."""
        raise NotImplementedError()



import asyncio
import logging

from opentrons.hardware_control.emulation.connection_handler import \
    ConnectionHandler
from opentrons.hardware_control.emulation.magdeck import MagDeckEmulator
from opentrons.hardware_control.emulation.parser import Parser
from opentrons.hardware_control.emulation.settings import Settings
from opentrons.hardware_control.emulation.tempdeck import TempDeckEmulator
from opentrons.hardware_control.emulation.thermocycler import ThermocyclerEmulator
from opentrons.hardware_control.emulation.smoothie import SmoothieEmulator

logger = logging.getLogger(__name__)


SMOOTHIE_PORT = 9996
THERMOCYCLER_PORT = 9997
TEMPDECK_PORT = 9998
MAGDECK_PORT = 9999


async def run_server(host: str, port: int, handler: ConnectionHandler) -> None:
    """Run a server."""
    server = await asyncio.start_server(handler, host, port)

    async with server:
        await server.serve_forever()


async def run() -> None:
    """Run the module emulators."""
    settings = Settings()
    host = settings.host

    await asyncio.gather(
        run_server(host=host,
                   port=MAGDECK_PORT,
                   handler=ConnectionHandler(MagDeckEmulator(parser=Parser()))),
        run_server(host=host,
                   port=TEMPDECK_PORT,
                   handler=ConnectionHandler(TempDeckEmulator(parser=Parser()))),
        run_server(host=host,
                   port=THERMOCYCLER_PORT,
                   handler=ConnectionHandler(ThermocyclerEmulator(parser=Parser()))),
        run_server(host=host,
                   port=SMOOTHIE_PORT,
                   handler=ConnectionHandler(
                       SmoothieEmulator(parser=Parser(), settings=settings.smoothie))
                   ),
    )


if __name__ == "__main__":
    logging.basicConfig(format='%(asctime)s:%(message)s', level=logging.DEBUG)
    asyncio.run(run())



from opentrons.config import advanced_settings as advs


def short_fixed_trash() -> bool:
    return advs.get_setting_with_env_overload('shortFixedTrash')


def calibrate_to_bottom() -> bool:
    return advs.get_setting_with_env_overload('calibrateToBottom')


def dots_deck_type() -> bool:
    return advs.get_setting_with_env_overload('deckCalibrationDots')


def disable_home_on_boot() -> bool:
    return advs.get_setting_with_env_overload('disableHomeOnBoot')


def use_protocol_api_v2() -> bool:
    return not advs.get_setting_with_env_overload('useLegacyInternals')


def use_old_aspiration_functions() -> bool:
    return advs.get_setting_with_env_overload('useOldAspirationFunctions')


def enable_door_safety_switch() -> bool:
    return advs.get_setting_with_env_overload('enableDoorSafetySwitch')


def enable_http_protocol_sessions() -> bool:
    return advs.get_setting_with_env_overload('enableHttpProtocolSessions')


def enable_fast_protocol_upload() -> bool:
    return advs.get_setting_with_env_overload('enableFastProtocolUpload')


def enable_protocol_engine() -> bool:
    """Get if the ProtocolEngine should be used to run protocol files."""

    return advs.get_setting_with_env_overload("enableProtocolEngine")



from __future__ import annotations
import enum
from math import sqrt, isclose
from typing import Any, NamedTuple, TYPE_CHECKING, Union

from .protocols.api_support.labware_like import LabwareLike

if TYPE_CHECKING:
    from .protocol_api.labware import Labware, Well
    from .protocols.geometry.module_geometry import ModuleGeometry


class PipetteNotAttachedError(KeyError):
    """ An error raised if a pipette is accessed that is not attached """
    pass


class Point(NamedTuple):
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Point):
            return False
        pairs = ((self.x, other.x), (self.y, other.y), (self.z, other.z))
        return all(isclose(s, o,
                           rel_tol=1e-05, abs_tol=1e-08) for s, o in pairs)

    def __add__(self, other: Any) -> 'Point':
        if not isinstance(other, Point):
            return NotImplemented
        return Point(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other: Any) -> 'Point':
        if not isinstance(other, Point):
            return NotImplemented
        return Point(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other: Union[int, float]) -> 'Point':
        if not isinstance(other, (float, int)):
            return NotImplemented
        return Point(self.x * other, self.y * other, self.z * other)

    def __rmul__(self, other: Union[int, float]) -> 'Point':
        if not isinstance(other, (float, int)):
            return NotImplemented
        return Point(self.x * other, self.y * other, self.z * other)

    def __abs__(self) -> 'Point':
        return Point(abs(self.x), abs(self.y), abs(self.z))

    def __str__(self):
        return '({}, {}, {})'.format(self.x, self.y, self.z)

    def magnitude_to(self, other: Any) -> float:
        if not isinstance(other, Point):
            return NotImplemented
        x_diff = self.x - other.x
        y_diff = self.y - other.y
        z_diff = self.z - other.z
        return sqrt(x_diff**2 + y_diff**2 + z_diff**2)


LocationLabware = Union['Labware', 'Well', str,
                        'ModuleGeometry', LabwareLike, None]


class Location:
    """ A location to target as a motion.

    The location contains a :py:class:`.Point` (in
    :ref:`protocol-api-deck-coords`) and possibly an associated
    :py:class:`.Labware` or :py:class:`.Well` instance.

    It should rarely be constructed directly by the user; rather, it is the
    return type of most :py:class:`.Well` accessors like :py:meth:`.Well.top`
    and is passed directly into a method like
    :py:meth:`InstrumentContext.aspirate`.

    .. warning::
       The :py:attr:`labware` attribute of this class is used by the protocol
       API internals to, among other things, determine safe heights to retract
       the instruments to when moving between locations. If constructing an
       instance of this class manually, be sure to either specify `None` as the
       labware (so the robot does its worst case retraction) or specify the
       correct labware for the :py:attr:`point` attribute.


    .. warning::
       The `==` operation compares both the position and associated labware.
       If you only need to compare locations, compare the :py:attr:`point`
       of each item.
    """

    def __init__(self, point: Point, labware: LocationLabware):
        self._point = point
        self._labware = LabwareLike(labware)

    @property
    def point(self) -> Point:
        return self._point

    @property
    def labware(self) -> LabwareLike:
        return self._labware

    def __iter__(self):
        """Iterable interface to support unpacking. Like a tuple."""
        return iter((self._point, self._labware,))

    def __eq__(self, other):
        return isinstance(other, Location) \
            and other._point == self._point \
            and other._labware == self._labware

    def move(self, point: Point) -> 'Location':
        """
        Alter the point stored in the location while preserving the labware.

        This returns a new Location and does not alter the current one. It
        should be used like

        .. code-block:: python

            >>> loc = Location(Point(1, 1, 1), None)
            >>> new_loc = loc.move(Point(1, 1, 1))
            >>> assert loc_2.point == Point(2, 2, 2)  # True
            >>> assert loc.point == Point(1, 1, 1)  # True

        """
        return Location(point=self.point + point,
                        labware=self._labware.object)

    def __repr__(self) -> str:
        return f"Location(point={repr(self._point)}, labware={self._labware})"


# TODO(mc, 2020-10-22): use MountType implementation for Mount
class Mount(enum.Enum):
    LEFT = enum.auto()
    RIGHT = enum.auto()

    def __str__(self):
        return self.name

    @classmethod
    def string_to_mount(cls, mount: str) -> 'Mount':
        if mount == 'right':
            return cls.RIGHT
        else:
            return cls.LEFT


class MountType(str, enum.Enum):
    LEFT = "left"
    RIGHT = "right"

    def other_mount(self) -> MountType:
        return MountType.LEFT if self is MountType.RIGHT else MountType.RIGHT

    def to_hw_mount(self) -> Mount:
        return Mount.LEFT if self is MountType.LEFT else Mount.RIGHT


# TODO(mc, 2020-11-09): this makes sense in shared-data or other common
# model library
# https://github.com/Opentrons/opentrons/pull/6943#discussion_r519029833
class DeckSlotName(int, enum.Enum):
    """Deck slot identifiers."""
    SLOT_1 = 1
    SLOT_2 = 2
    SLOT_3 = 3
    SLOT_4 = 4
    SLOT_5 = 5
    SLOT_6 = 6
    SLOT_7 = 7
    SLOT_8 = 8
    SLOT_9 = 9
    SLOT_10 = 10
    SLOT_11 = 11
    FIXED_TRASH = 12

    @classmethod
    def from_primitive(cls, value: DeckLocation) -> DeckSlotName:
        int_val = int(value)
        return cls(int_val)

    def __str__(self):
        """Stringify to a simple integer string."""
        return str(self.value)


class TransferTipPolicy(enum.Enum):
    ONCE = enum.auto()
    NEVER = enum.auto()
    ALWAYS = enum.auto()


DeckLocation = Union[int, str]



import asyncio
import logging
import os
from pathlib import Path
from glob import glob
from typing import Any, Dict, Tuple, Optional
from .types import UpdateError
from .mod_abc import AbstractModule

log = logging.getLogger(__name__)


async def update_firmware(
        module: AbstractModule,
        firmware_file: str,
        loop: Optional[asyncio.AbstractEventLoop]):
    """ Apply update of given firmware file to given module.

        raises an UpdateError with the reason for the failure.
    """
    flash_port = await module.prep_for_update()
    kwargs: Dict[str, Any] = {
        'stdout': asyncio.subprocess.PIPE,
        'stderr': asyncio.subprocess.PIPE,
        'loop': loop
    }
    successful, res = await module.bootloader()(
        flash_port, firmware_file, kwargs)
    if not successful:
        log.info(f'Bootloader reponse: {res}')
        raise UpdateError(res)


async def find_bootloader_port():
    """
    Finds the port of an Opentrons Module that has entered its bootloader.
    The bootloader port shows up as 'ot_module_(avrdude|samba)_bootloader'
    on the pi; return found port.
    """

    for attempt in range(3):
        bootloader_ports = glob('/dev/ot_module_*_bootloader*')
        if bootloader_ports:
            if len(bootloader_ports) == 1:
                log.info(f'Found bootloader at port {bootloader_ports[0]}')
                return bootloader_ports[0]
            elif len(bootloader_ports) > 1:
                raise OSError('Multiple new bootloader ports'
                              'found on mode switch')
        await asyncio.sleep(2)
    raise Exception("No ot_module bootloaders found in /dev. Try again")


async def upload_via_avrdude(port: str,
                             firmware_file_path: str,
                             kwargs: Dict[str, Any]) -> Tuple[bool, str]:
    """
    Run firmware upload command for hardware module with avrdude bootloader.

    Returns tuple of success boolean and message from bootloader.
    """
    # avrdude_options
    PART_NO = 'atmega32u4'
    PROGRAMMER_ID = 'avr109'
    BAUDRATE = '57600'

    config_file_path = Path('/etc/avrdude.conf')
    proc = await asyncio.create_subprocess_exec(
        'avrdude', '-C{}'.format(config_file_path), '-v',
        '-p{}'.format(PART_NO),
        '-c{}'.format(PROGRAMMER_ID),
        '-P{}'.format(port),
        '-b{}'.format(BAUDRATE), '-D',
        '-Uflash:w:{}:i'.format(firmware_file_path),
        **kwargs)
    await proc.wait()

    _result = await proc.communicate()
    result = _result[1].decode()
    avrdude_res = _format_avrdude_response(result)
    if avrdude_res[0]:
        log.debug(result)
    else:
        log.error("Failed to update module firmware for {}: {}"
                  .format(port, avrdude_res[1]))
    return avrdude_res


def _format_avrdude_response(raw_response: str) -> Tuple[bool, str]:
    avrdude_log = ''
    for line in raw_response.splitlines():
        if 'avrdude:' in line and line != raw_response.splitlines()[1]:
            avrdude_log += line.lstrip('avrdude:') + '..'
            if 'flash verified' in line:
                return True, line.lstrip('avrdude: ')
    return False, avrdude_log


async def upload_via_bossa(port: str,
                           firmware_file_path: str,
                           kwargs: Dict[str, Any]) -> Tuple[bool, str]:
    """
    Run firmware upload command for hardware module with SAMBA bootloader.

    Returns tuple of success boolean and message from bootloader.
    """
    # bossac -p/dev/ttyACM1 -e -w -v -R --offset=0x2000
    #   modules/thermo-cycler/production/firmware/thermo-cycler-arduino.ino.bin
    # NOTE: bossac cannot traverse symlinks to port,
    # so we resolve to real path
    resolved_symlink = os.path.realpath(port)
    log.info(
        f"device at symlinked port: {port} "
        f"resolved to path: {resolved_symlink}")
    bossa_args = ['bossac', f'-p{resolved_symlink}',
                  '-e', '-w', '-v', '-R',
                  '--offset=0x2000', f'{firmware_file_path}']

    proc = await asyncio.create_subprocess_exec(*bossa_args, **kwargs)
    stdout, stderr = await proc.communicate()
    res = stdout.decode()
    if "Verify successful" in res:
        log.debug(res)
        return True, res
    elif stderr:
        log.error(f"Failed to update module firmware for {port}: {res}")
        log.error(f"Error given: {stderr.decode()}")
        return False, res
    return False, ''



from abc import ABC, abstractmethod

from opentrons.types import Location, Point
from opentrons_shared_data.labware.dev_types import LabwareDefinition


class AbstractLabwareGeometry(ABC):
    """Convenience interface for accessing geometry properties of a labware."""

    @property
    @abstractmethod
    def parent(self) -> Location:
        ...

    @property
    @abstractmethod
    def offset(self) -> Point:
        ...

    @property
    @abstractmethod
    def x_dimension(self) -> float:
        ...

    @property
    @abstractmethod
    def y_dimension(self) -> float:
        ...

    @property
    @abstractmethod
    def z_dimension(self) -> float:
        ...


class LabwareGeometry(AbstractLabwareGeometry):

    def __init__(self,
                 definition: LabwareDefinition,
                 parent: Location):
        """Constructor"""
        self._parent = parent
        offset = definition['cornerOffsetFromSlot']
        self._offset = Point(
            offset['x'],
            offset['y'],
            offset['z']
        ) + parent.point

        dimensions = definition['dimensions']
        self._x_dimension = dimensions['xDimension']
        self._y_dimension = dimensions['yDimension']
        self._z_dimension = dimensions['zDimension']

    @property
    def parent(self) -> Location:
        return self._parent

    @property
    def offset(self) -> Point:
        return self._offset

    @property
    def x_dimension(self) -> float:
        return self._x_dimension

    @property
    def y_dimension(self) -> float:
        return self._y_dimension

    @property
    def z_dimension(self) -> float:
        return self._z_dimension
